/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <sched.h>
#include <unistd.h>
#include <string.h>

#include "mars/mars_workload_queue.h"
#include "mars/mars_kernel_types.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

int workload_queue_initialize(struct mars_workload_queue *queue)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);

	int block;
	int index;

	queue->header.queue_ea = (uint64_t)(uintptr_t)queue;
	queue->header.context_ea = (uint64_t)(uintptr_t)&queue->context;
	queue->header.flag = MARS_FLAG_NONE;

	/* initialize workload queue blocks */
	for (block = 0; block < MARS_WORKLOAD_NUM_BLOCKS; block++) {
		struct mars_workload_queue_block *p = &queue->block[block];

		mars_mutex_initialize((struct mars_mutex *)p);
		mars_mutex_lock((struct mars_mutex *)p);

		for (index = 0; index < MARS_WORKLOAD_PER_BLOCK; index++) {
			p->bits[index].type = MARS_WORKLOAD_TYPE_NONE;
			p->bits[index].state = MARS_WORKLOAD_STATE_NONE;
			p->bits[index].priority = MARS_WORKLOAD_PRIORITY_MIN;
			p->bits[index].counter = MARS_WORKLOAD_COUNTER_MIN;
			p->bits[index].signal = MARS_WORKLOAD_SIGNAL_OFF;
			p->bits[index].wait = MARS_WORKLOAD_ID_NONE;
		}

		mars_mutex_unlock((struct mars_mutex *)p);
	}

	return MARS_SUCCESS;
}

int workload_queue_finalize(struct mars_workload_queue *queue)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);

	int block;
	int index;

	uint16_t id = 0;

	/* check for any tasks left in workload queue */
	while (id < MARS_WORKLOAD_MAX) {
		block = id / MARS_WORKLOAD_PER_BLOCK;
		index = id % MARS_WORKLOAD_PER_BLOCK;

		if (queue->block[block].bits[index].state !=
			MARS_WORKLOAD_STATE_NONE)
			break;

		id++;
	}

	/* found some task left in workload queue */
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_STATE);

	queue->header.flag = MARS_FLAG_EXIT;

	return MARS_SUCCESS;
}

int workload_queue_add_begin(struct mars_workload_queue *queue, uint16_t *id,
				uint8_t type,
				struct mars_workload_context **workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(type == MARS_WORKLOAD_TYPE_TASK, MARS_ERROR_PARAMS);

	int block = 0;
	int index = 0;

	*id = 0;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	while (queue->block[block].bits[index].state !=
		MARS_WORKLOAD_STATE_NONE) {
		(*id)++;
		index++;
		if (index == MARS_WORKLOAD_PER_BLOCK) {
			index = 0;

			mars_mutex_unlock(
				(struct mars_mutex *)&queue->block[block]);

			if (++block == MARS_WORKLOAD_NUM_BLOCKS)
				return MARS_ERROR_LIMIT;

			mars_mutex_lock(
				(struct mars_mutex *)&queue->block[block]);
		}
	}

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_NONE,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set type and set state to adding */
	queue->block[block].bits[index].type = type;
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_ADDING;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[*id];

	return MARS_SUCCESS;
}

int workload_queue_add_end(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_ADDING,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* reset workload queue bits and set state to finished state */
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_FINISHED;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int workload_queue_add_cancel(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_ADDING,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set state back to none state */
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_NONE;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int workload_queue_remove_begin(struct mars_workload_queue *queue, uint16_t id,
				struct mars_workload_context **workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_FINISHED,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set state to removing */
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_REMOVING;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int workload_queue_remove_end(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_REMOVING,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set state to none */
	queue->block[block].bits[index].type = MARS_WORKLOAD_TYPE_NONE;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int workload_queue_remove_cancel(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_REMOVING,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set state back to finished */
	queue->block[block].bits[index].type = MARS_WORKLOAD_STATE_FINISHED;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int workload_queue_schedule_begin(struct mars_workload_queue *queue,
				uint16_t id, uint8_t priority,
				struct mars_workload_context **workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_FINISHED,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* reset workload queue bits and set state to scheduling */
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_SCHEDULING;
	queue->block[block].bits[index].priority = priority;
	queue->block[block].bits[index].counter = MARS_WORKLOAD_COUNTER_MIN;
	queue->block[block].bits[index].signal = MARS_WORKLOAD_SIGNAL_OFF;
	queue->block[block].bits[index].wait = MARS_WORKLOAD_ID_NONE;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int workload_queue_schedule_end(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_SCHEDULING,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set state to ready */
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_READY;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int workload_queue_schedule_cancel(struct mars_workload_queue *queue,
				uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state ==
		MARS_WORKLOAD_STATE_SCHEDULING,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	/* set state back to finished */
	queue->block[block].bits[index].state = MARS_WORKLOAD_STATE_FINISHED;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int workload_queue_wait(struct mars_workload_queue *queue, uint16_t id,
			struct mars_workload_context **workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	while (queue->block[block].bits[index].state !=
		MARS_WORKLOAD_STATE_FINISHED) {
		MARS_CHECK_RET(queue->block[block].bits[index].state !=
			MARS_WORKLOAD_STATE_NONE,
			MARS_ERROR_STATE);
		sched_yield();
	}

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int workload_queue_try_wait(struct mars_workload_queue *queue, uint16_t id,
			struct mars_workload_context **workload)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	MARS_CHECK_RET(queue->block[block].bits[index].state !=
		MARS_WORKLOAD_STATE_NONE,
		MARS_ERROR_STATE);

	if (queue->block[block].bits[index].state !=
		MARS_WORKLOAD_STATE_FINISHED)
		return MARS_ERROR_BUSY;

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int workload_queue_signal_send(struct mars_workload_queue *queue, uint16_t id)
{
	MARS_CHECK_RET(queue, MARS_ERROR_NULL);
	MARS_CHECK_RET(id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int block = id / MARS_WORKLOAD_PER_BLOCK;
	int index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	MARS_CHECK_CLEANUP_RET(queue->block[block].bits[index].state !=
		MARS_WORKLOAD_STATE_NONE,
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]),
		MARS_ERROR_STATE);

	queue->block[block].bits[index].signal = MARS_WORKLOAD_SIGNAL_ON;

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}
