/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "mars/mars_kernel.h"

#define MARS_WORKLOAD_RESERVED_NONE	0
#define MARS_WORKLOAD_RESERVED		1

struct mars_workload_queue_header queue_header;
struct mars_workload_queue_block queue_block;
struct mars_workload_context workload;
uint8_t workload_type;
uint8_t workload_state;
uint16_t workload_index;
uint64_t workload_ea;

int reserve_workload(void);
void release_workload(void);

static inline void get_block(int block, struct mars_workload_queue_block *dst)
{
	mars_dma_get_and_wait(dst,
			queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block,
			sizeof(struct mars_workload_queue_block),
			MARS_DMA_TAG);
}

static int search_block(int block)
{
	int i;
	int index = -1;
	int max_count = -1;
	int max_priority = -1;

	/* search through workload queue for next workload to run
	 * while incrementing wait counter for all waiting workloads
	 * and pick the workload that has been waiting the longest
	 */
	for (i = 0; i < MARS_WORKLOAD_PER_BLOCK; i++) {
		struct mars_workload_queue_block_bits *bits
			= &queue_block.bits[i];

		switch (bits->state) {
		case MARS_WORKLOAD_STATE_READY:
			/* priority greater than max priority so select */
			if ((int)bits->priority > max_priority) {
				index = i;
				max_count = bits->counter;
				max_priority = bits->priority;
			/* priority equal and wait counter greater so select */
			} else if ((int)bits->priority == max_priority &&
				(int)bits->counter > max_count) {
				index = i;
				max_count = bits->counter;
			}
			/* increment wait counter without overflowing */
			if (bits->counter < MARS_WORKLOAD_COUNTER_MAX)
				bits->counter++;
			break;
		case MARS_WORKLOAD_STATE_WAITING:
			/* waiting for workload to finish so check status */
			if (bits->wait != MARS_WORKLOAD_ID_NONE) {
				struct mars_workload_queue_block wait_block;
				struct mars_workload_queue_block *p_wait_block;

				int bl = bits->wait / MARS_WORKLOAD_PER_BLOCK;
				int id = bits->wait % MARS_WORKLOAD_PER_BLOCK;

				/* check if workload id is in the same block */
				if (block != bl) {
					/* fetch the necessary block */
					get_block(bl, &wait_block);
					/* set pointer to check fetched block */
					p_wait_block = &wait_block;
				} else {
					/* set pointer to check current block */
					p_wait_block = &queue_block;
				}

				/* check if workload is finished and reset */
				if (p_wait_block->bits[id].state ==
					MARS_WORKLOAD_STATE_FINISHED) {
					bits->wait = MARS_WORKLOAD_ID_NONE;
					bits->state = MARS_WORKLOAD_STATE_READY;
				}
			/* waiting for signal so check signal bit and reset */
			} else if (bits->signal == MARS_WORKLOAD_SIGNAL_ON) {
				bits->signal = MARS_WORKLOAD_SIGNAL_OFF;
				bits->state = MARS_WORKLOAD_STATE_READY;
				i--;
			}
			break;
		default:
			break;
		}
	}

	/* returns -1 if no runnable workload found */
	return index;
}

static int reserve_block(int block)
{
	int index;
	uint64_t block_ea = queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* set the workload index */
	index = search_block(block);
	if (index >= 0) {
		/* update the current state of the workload */
		queue_block.bits[index].state = MARS_WORKLOAD_STATE_RUNNING;

		/* reset the counter for reserved workload */
		queue_block.bits[index].counter = MARS_WORKLOAD_COUNTER_MIN;
	}

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return index;
}

static void release_block(int block, int index)
{
	uint64_t block_ea = queue_header.queue_ea +
			offsetof(struct mars_workload_queue, block) +
			sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* update current workload state in workload queue block */
	queue_block.bits[index].state = workload_state;

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);
}

int reserve_workload(void)
{
	int block;
	int index = -1;

	/* search workload queue blocks until runnable workload reserved */
	for (block = 0; block < MARS_WORKLOAD_NUM_BLOCKS; block++) {
		index = reserve_block(block);
		if (index >= 0)
			break;
	}

	/* no runnable workload found */
	if (index < 0)
		return MARS_WORKLOAD_RESERVED_NONE;

	/* set global workload info based on workload block and index */
	workload_index = MARS_WORKLOAD_PER_BLOCK * block + index;
	workload_type = queue_block.bits[index].type;
	workload_ea = queue_header.context_ea +
			workload_index * sizeof(struct mars_workload_context);

	/* dma the workload context code into LS from main memory */
	mars_dma_get_and_wait((void *)&workload,
				workload_ea,
				sizeof(struct mars_workload_context),
				MARS_DMA_TAG);

	return MARS_WORKLOAD_RESERVED;
}

void release_workload(void)
{
	int block = workload_index / MARS_WORKLOAD_PER_BLOCK;
	int index = workload_index % MARS_WORKLOAD_PER_BLOCK;

	/* dma updated workload context back to main memory if not finished */
	if (workload_state != MARS_WORKLOAD_STATE_FINISHED) {
		mars_dma_put_and_wait((void *)&workload,
					workload_ea,
					sizeof(struct mars_workload_context),
					MARS_DMA_TAG);
	}

	/* release block reservation */
	release_block(block, index);
}

int scheduler(void)
{
	int status;

	/* dma in queue header */
	mars_dma_get_and_wait(&queue_header,
			kernel_params.workload_queue_ea +
			offsetof(struct mars_workload_queue, header),
			sizeof(struct mars_workload_queue_header),
			MARS_DMA_TAG);

	/* return exit status if exit flag is set from host */
	if (queue_header.flag == MARS_FLAG_EXIT)
		return MARS_KERNEL_STATUS_EXIT;

	/* reserve next workload to run or return idle status if none found */
	status = reserve_workload();

	/* return idle status if no workload was reserved */
	if (status == MARS_WORKLOAD_RESERVED_NONE)
		return MARS_KERNEL_STATUS_IDLE;

	/* run the reserved workload */
	workload_run();

	/* release reservation of current workload */
	release_workload();

	return MARS_KERNEL_STATUS_BUSY;
}
